import pathlib

from django.core.management.base import BaseCommand, CommandError
from strawberry import Schema
from strawberry.printer import print_schema
from strawberry.utils.importer import import_module_symbol


class Command(BaseCommand):
    help = "Export the graphql schema"

    def add_arguments(self, parser):
        parser.add_argument("schema", nargs=1, type=str, help="The schema location")
        parser.add_argument(
            "--path",
            nargs="?",
            type=str,
            help="Optional path to export",
        )

    def handle(self, *args, **options):
        try:
            schema_symbol = import_module_symbol(
                options["schema"][0],
                default_symbol_name="schema",
            )
        except (ImportError, AttributeError) as e:
            raise CommandError(str(e)) from e

        if not isinstance(schema_symbol, Schema):
            raise CommandError("The `schema` must be an instance of strawberry.Schema")

        schema_output = print_schema(schema_symbol)
        path = options.get("path")
        if path:
            pathlib.Path(path).write_text(schema_output, encoding="utf-8")
        else:
            self.stdout.write(schema_output)



import django
from django.db import (
    DEFAULT_DB_ALIAS,
    NotSupportedError,
    connections,
)
from django.db.models import Q, Window
from django.db.models.fields import related_descriptors
from django.db.models.functions import RowNumber
from django.db.models.lookups import GreaterThan, LessThanOrEqual
from django.db.models.sql import Query
from django.db.models.sql.constants import INNER
from django.db.models.sql.where import AND


def apply_pagination_fix():
    """Apply pagination fix for Django 5.1 or older.

    This is based on the fix in this patch, which is going to be included in Django 5.2:
    https://code.djangoproject.com/ticket/35677#comment:9

    If can safely be removed when Django 5.2 is the minimum version we support
    """
    if django.VERSION >= (5, 2):
        return

    # This is a copy of the function, exactly as it exists on Django 4.2, 5.0 and 5.1
    # (there are no differences in this function between these versions)
    def _filter_prefetch_queryset(queryset, field_name, instances):
        predicate = Q(**{f"{field_name}__in": instances})
        db = queryset._db or DEFAULT_DB_ALIAS
        if queryset.query.is_sliced:
            if not connections[db].features.supports_over_clause:
                raise NotSupportedError(
                    "Prefetching from a limited queryset is only supported on backends "
                    "that support window functions."
                )
            low_mark, high_mark = queryset.query.low_mark, queryset.query.high_mark
            order_by = [
                expr for expr, _ in queryset.query.get_compiler(using=db).get_order_by()
            ]
            window = Window(RowNumber(), partition_by=field_name, order_by=order_by)
            predicate &= GreaterThan(window, low_mark)
            if high_mark is not None:
                predicate &= LessThanOrEqual(window, high_mark)
            queryset.query.clear_limits()

        # >> ORIGINAL CODE
        # return queryset.filter(predicate)  # noqa: ERA001
        # << ORIGINAL CODE
        # >> PATCHED CODE
        queryset.query.add_q(predicate, reuse_all_aliases=True)
        return queryset
        # << PATCHED CODE

    related_descriptors._filter_prefetch_queryset = _filter_prefetch_queryset  # type: ignore

    # This is a copy of the function, exactly as it exists on Django 4.2, 5.0 and 5.1
    # (there are no differences in this function between these versions)
    def add_q(self, q_object, reuse_all_aliases=False):
        existing_inner = {
            a for a in self.alias_map if self.alias_map[a].join_type == INNER
        }
        # >> ORIGINAL CODE
        # clause, _ = self._add_q(q_object, self.used_aliases)  # noqa: ERA001
        # << ORIGINAL CODE
        # >> PATCHED CODE
        if reuse_all_aliases:  # noqa: SIM108
            can_reuse = set(self.alias_map)
        else:
            can_reuse = self.used_aliases
        clause, _ = self._add_q(q_object, can_reuse)
        # << PATCHED CODE
        if clause:
            self.where.add(clause, AND)
        self.demote_joins(existing_inner)

    Query.add_q = add_q



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Any, TypeVar

from django.db.models import Model
from django.db.models.query import QuerySet

if TYPE_CHECKING:
    from strawberry import Info

_M = TypeVar("_M", bound=Model)

CONFIG_KEY = "_strawberry_django_config"


@dataclasses.dataclass
class StrawberryDjangoQuerySetConfig:
    optimized: bool = False
    type_get_queryset_did_run: bool = False


def get_queryset_config(queryset: QuerySet) -> StrawberryDjangoQuerySetConfig:
    return getattr(queryset, CONFIG_KEY, None) or StrawberryDjangoQuerySetConfig()


def run_type_get_queryset(
    qs: QuerySet[_M],
    origin: Any,
    info: Info | None = None,
) -> QuerySet[_M]:
    config = get_queryset_config(qs)
    get_queryset = getattr(origin, "get_queryset", None)

    if get_queryset and not config.type_get_queryset_did_run:
        qs = get_queryset(qs, info)
        new_config = get_queryset_config(qs)
        new_config.type_get_queryset_did_run = True

    return qs


_original_clone = QuerySet._clone  # type: ignore


def _qs_clone(self):
    config = get_queryset_config(self)
    cloned = _original_clone(self)
    setattr(cloned, CONFIG_KEY, dataclasses.replace(config))
    return cloned


# Monkey patch the QuerySet._clone method to make sure our config is copied
# to the new QuerySet instance once it is cloned.
QuerySet._clone = _qs_clone  # type: ignore



from __future__ import annotations

import dataclasses
import enum
from typing import (
    TYPE_CHECKING,
    Optional,
    TypeVar,
    cast,
)

import strawberry
from django.db.models import F, OrderBy, QuerySet
from graphql.language.ast import ObjectValueNode
from strawberry import UNSET
from strawberry.types import has_object_definition
from strawberry.types.base import WithStrawberryObjectDefinition
from strawberry.types.field import StrawberryField, field
from strawberry.types.unset import UnsetType
from strawberry.utils.str_converters import to_camel_case
from typing_extensions import Self, dataclass_transform

from strawberry_django.fields.base import StrawberryDjangoFieldBase
from strawberry_django.fields.filter_order import (
    WITH_NONE_META,
    FilterOrderField,
    FilterOrderFieldResolver,
)
from strawberry_django.utils.typing import is_auto

from .arguments import argument

if TYPE_CHECKING:
    from collections.abc import Callable, Collection, Sequence

    from django.db.models import Model
    from strawberry.types import Info
    from strawberry.types.arguments import StrawberryArgument

_T = TypeVar("_T")
_QS = TypeVar("_QS", bound="QuerySet")
_SFT = TypeVar("_SFT", bound=StrawberryField)

ORDER_ARG = "order"


@dataclasses.dataclass
class OrderSequence:
    seq: int = 0
    children: dict[str, OrderSequence] | None = None

    @classmethod
    def get_graphql_name(cls, info: Info | None, field: StrawberryField) -> str:
        if info is None:
            if field.graphql_name:
                return field.graphql_name

            return to_camel_case(field.python_name)

        return info.schema.config.name_converter.get_graphql_name(field)

    @classmethod
    def sorted(
        cls,
        info: Info | None,
        sequence: dict[str, OrderSequence] | None,
        fields: list[_SFT],
    ) -> list[_SFT]:
        if info is None:
            return fields

        sequence = sequence or {}

        def sort_key(f: _SFT) -> int:
            if not (seq := sequence.get(cls.get_graphql_name(info, f))):
                return 0
            return seq.seq

        return sorted(fields, key=sort_key)


@strawberry.enum
class Ordering(enum.Enum):
    ASC = "ASC"
    ASC_NULLS_FIRST = "ASC_NULLS_FIRST"
    ASC_NULLS_LAST = "ASC_NULLS_LAST"
    DESC = "DESC"
    DESC_NULLS_FIRST = "DESC_NULLS_FIRST"
    DESC_NULLS_LAST = "DESC_NULLS_LAST"

    def resolve(self, value: str) -> OrderBy:
        nulls_first = True if "NULLS_FIRST" in self.name else None
        nulls_last = True if "NULLS_LAST" in self.name else None
        if "ASC" in self.name:
            return F(value).asc(nulls_first=nulls_first, nulls_last=nulls_last)
        return F(value).desc(nulls_first=nulls_first, nulls_last=nulls_last)


def process_order(
    order: WithStrawberryObjectDefinition,
    info: Info | None,
    queryset: _QS,
    *,
    sequence: dict[str, OrderSequence] | None = None,
    prefix: str = "",
    skip_object_order_method: bool = False,
) -> tuple[_QS, Collection[F | OrderBy | str]]:
    sequence = sequence or {}
    args = []

    if not skip_object_order_method and isinstance(
        order_method := getattr(order, "order", None),
        FilterOrderFieldResolver,
    ):
        return order_method(
            order, info, queryset=queryset, prefix=prefix, sequence=sequence
        )

    for f in OrderSequence.sorted(
        info, sequence, order.__strawberry_definition__.fields
    ):
        f_value = getattr(order, f.name, UNSET)
        if f_value is UNSET or (f_value is None and not f.metadata.get(WITH_NONE_META)):
            continue

        if isinstance(f, FilterOrderField) and f.base_resolver:
            res = f.base_resolver(
                order,
                info,
                value=f_value,
                queryset=queryset,
                prefix=prefix,
                sequence=(
                    (seq := sequence.get(OrderSequence.get_graphql_name(info, f)))
                    and seq.children
                ),
            )
            if isinstance(res, tuple):
                queryset, subargs = res
            else:
                subargs = res
            args.extend(subargs)
        elif isinstance(f_value, Ordering):
            args.append(f_value.resolve(f"{prefix}{f.name}"))
        else:
            queryset, subargs = process_order(
                f_value,
                info,
                queryset,
                prefix=f"{prefix}{f.name}__",
                sequence=(
                    (seq := sequence.get(OrderSequence.get_graphql_name(info, f)))
                    and seq.children
                ),
            )
            args.extend(subargs)

    return queryset, args


def apply(
    order: object | None,
    queryset: _QS,
    info: Info | None = None,
) -> _QS:
    if order in (None, strawberry.UNSET) or not has_object_definition(order):  # noqa: PLR6201
        return queryset

    sequence: dict[str, OrderSequence] = {}
    if info is not None and info._raw_info.field_nodes:
        field_node = info._raw_info.field_nodes[0]
        for arg in field_node.arguments:
            if arg.name.value != ORDER_ARG or not isinstance(
                arg.value, ObjectValueNode
            ):
                continue

            def parse_and_fill(field: ObjectValueNode, seq: dict[str, OrderSequence]):
                for i, f in enumerate(field.fields):
                    f_sequence: dict[str, OrderSequence] = {}
                    if isinstance(f.value, ObjectValueNode):
                        parse_and_fill(f.value, f_sequence)

                    seq[f.name.value] = OrderSequence(seq=i, children=f_sequence)

            parse_and_fill(arg.value, sequence)

    queryset, args = process_order(
        cast("WithStrawberryObjectDefinition", order), info, queryset, sequence=sequence
    )
    if not args:
        return queryset
    return queryset.order_by(*args)


class StrawberryDjangoFieldOrdering(StrawberryDjangoFieldBase):
    def __init__(self, order: type | UnsetType | None = UNSET, **kwargs):
        if order and not has_object_definition(order):
            raise TypeError("order needs to be a strawberry type")

        self.order = order
        super().__init__(**kwargs)

    def __copy__(self) -> Self:
        new_field = super().__copy__()
        new_field.order = self.order
        return new_field

    @property
    def arguments(self) -> list[StrawberryArgument]:
        arguments = []
        if self.base_resolver is None and self.is_list:
            order = self.get_order()
            if order and order is not UNSET:
                arguments.append(argument("order", order, is_optional=True))
        return super().arguments + arguments

    @arguments.setter
    def arguments(self, value: list[StrawberryArgument]):
        args_prop = super(StrawberryDjangoFieldOrdering, self.__class__).arguments
        return args_prop.fset(self, value)  # type: ignore

    def get_order(self) -> type[WithStrawberryObjectDefinition] | None:
        order = self.order
        if order is None:
            return None

        if isinstance(order, UnsetType):
            django_type = self.django_type
            order = (
                django_type.__strawberry_django_definition__.order
                if django_type is not None
                else None
            )

        return order if order is not UNSET else None

    def get_queryset(
        self,
        queryset: _QS,
        info: Info,
        *,
        order: WithStrawberryObjectDefinition | None = None,
        **kwargs,
    ) -> _QS:
        queryset = super().get_queryset(queryset, info, **kwargs)
        return apply(order, queryset, info=info)


@dataclass_transform(
    order_default=True,
    field_specifiers=(
        StrawberryField,
        field,
    ),
)
def order(
    model: type[Model],
    *,
    name: str | None = None,
    description: str | None = None,
    directives: Sequence[object] | None = (),
) -> Callable[[_T], _T]:
    def wrapper(cls):
        try:
            cls.__annotations__  # noqa: B018
        except AttributeError:
            # FIXME: Manual creation for python 3.9 (remove when 3.9 is dropped)
            cls.__annotations__ = {}

        for fname, type_ in cls.__annotations__.items():
            if is_auto(type_):
                type_ = Ordering  # noqa: PLW2901

            cls.__annotations__[fname] = Optional[type_]

            field_ = cls.__dict__.get(fname)
            if not isinstance(field_, StrawberryField):
                setattr(cls, fname, UNSET)

        return strawberry.input(
            cls,
            name=name,
            description=description,
            directives=directives,
        )

    return wrapper



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Annotated, Any, TypeVar, Union

import strawberry
from django.core.exceptions import (
    NON_FIELD_ERRORS,
    ObjectDoesNotExist,
    PermissionDenied,
    ValidationError,
)
from django.db import models, transaction
from strawberry import UNSET, relay
from strawberry.annotation import StrawberryAnnotation
from strawberry.types.field import UNRESOLVED
from strawberry.utils.str_converters import capitalize_first, to_camel_case

from strawberry_django.arguments import argument
from strawberry_django.fields.field import (
    StrawberryDjangoFieldBase,
    StrawberryDjangoFieldFilters,
)
from strawberry_django.fields.types import OperationInfo, OperationMessage
from strawberry_django.optimizer import DjangoOptimizerExtension, optimize
from strawberry_django.permissions import filter_with_perms, get_with_perms
from strawberry_django.resolvers import django_resolver
from strawberry_django.settings import strawberry_django_settings
from strawberry_django.utils.inspect import get_possible_types

from . import resolvers

if TYPE_CHECKING:
    from collections.abc import Iterable

    from graphql.pyutils import AwaitableOrValue
    from strawberry.types import Info
    from strawberry.types.arguments import StrawberryArgument
    from strawberry.types.base import (
        StrawberryObjectDefinition,
        StrawberryType,
        WithStrawberryObjectDefinition,
    )
    from typing_extensions import Literal, Self

    from .types import FullCleanOptions

_T = TypeVar("_T", bound="models.Model | list[models.Model]")


def _get_validaton_error_message(error: ValidationError):
    if not error.message:
        return "Unknown error"

    return error.message % error.params if error.params else error.message


def _get_validation_errors(error: Exception):
    if isinstance(error, PermissionDenied):
        kind = OperationMessage.Kind.PERMISSION
    elif isinstance(error, ValidationError):
        kind = OperationMessage.Kind.VALIDATION
    elif isinstance(error, ObjectDoesNotExist):
        kind = OperationMessage.Kind.ERROR
    else:
        kind = OperationMessage.Kind.ERROR

    if isinstance(error, ValidationError) and hasattr(error, "error_dict"):
        # convert field errors
        for field, field_errors in (error.error_dict or {}).items():
            for e in field_errors:
                yield OperationMessage(
                    kind=kind,
                    field=to_camel_case(field) if field != NON_FIELD_ERRORS else None,
                    message=_get_validaton_error_message(e),
                    code=getattr(e, "code", None),
                )
    elif isinstance(error, ValidationError) and hasattr(error, "error_list"):
        # convert non-field errors
        for e in error.error_list or []:
            yield OperationMessage(
                kind=kind,
                message=_get_validaton_error_message(e),
                code=getattr(error, "code", None),
            )
    else:
        msg = getattr(error, "msg", None)
        if msg is None:
            msg = str(error)

        yield OperationMessage(
            kind=kind,
            message=msg,
            code=getattr(error, "code", None),
        )


def _handle_exception(error: Exception):
    if isinstance(error, (ValidationError, PermissionDenied, ObjectDoesNotExist)):
        return OperationInfo(
            messages=list(_get_validation_errors(error)),
        )

    raise error


class DjangoMutationBase(StrawberryDjangoFieldBase):
    def __init__(
        self,
        *args,
        handle_django_errors: bool | None = None,
        **kwargs,
    ):
        self._resolved_return_type: bool = False

        if handle_django_errors is None:
            settings = strawberry_django_settings()
            handle_django_errors = settings["MUTATIONS_DEFAULT_HANDLE_ERRORS"]
        self.handle_errors = handle_django_errors

        super().__init__(*args, **kwargs)

    def __copy__(self) -> Self:
        new_field = super().__copy__()
        new_field.handle_errors = self.handle_errors
        return new_field

    def resolve_type(
        self,
        *,
        type_definition: StrawberryObjectDefinition | None = None,
    ) -> (
        StrawberryType | type[WithStrawberryObjectDefinition] | Literal[UNRESOLVED]  # type: ignore
    ):
        resolved = super().resolve_type(type_definition=type_definition)
        if resolved is UNRESOLVED:
            return resolved

        if self.handle_errors and not self._resolved_return_type:
            types_ = tuple(get_possible_types(resolved))
            if OperationInfo not in types_:
                types_ = (*types_, OperationInfo)

                name = capitalize_first(to_camel_case(self.python_name))
                resolved = Annotated[
                    Union[types_],
                    strawberry.union(f"{name}Payload"),
                ]
                self.type_annotation = StrawberryAnnotation(
                    resolved,
                    namespace=getattr(self.type_annotation, "namespace", None),
                )

            self._resolved_return_type = True

        return resolved

    def get_result(
        self,
        source: Any,
        info: Info | None,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> AwaitableOrValue[Any]:
        if not self.handle_errors:
            return self.resolver(source, info, args, kwargs)

        # TODO: Any other exception types that we should capture here?
        try:
            resolved = self.resolver(source, info, args, kwargs)
        except Exception as e:  # noqa: BLE001
            return _handle_exception(e)

        if inspect.isawaitable(resolved):

            async def async_resolver():
                try:
                    return await resolved
                except Exception as e:  # noqa: BLE001
                    return _handle_exception(e)

            return async_resolver()

        return resolved


class DjangoMutationCUD(DjangoMutationBase):
    def __init__(
        self,
        input_type: type | None = None,
        full_clean: bool | FullCleanOptions = True,
        argument_name: str | None = None,
        key_attr: str | None = None,
        **kwargs,
    ):
        self.full_clean = full_clean
        self.input_type = input_type

        if key_attr is None:
            settings = strawberry_django_settings()
            key_attr = settings["DEFAULT_PK_FIELD_NAME"]

        self.key_attr = key_attr

        if argument_name is None:
            settings = strawberry_django_settings()
            argument_name = settings["MUTATIONS_DEFAULT_ARGUMENT_NAME"]
        self.argument_name = argument_name

        super().__init__(**kwargs)

    def __copy__(self) -> Self:
        new_field = super().__copy__()
        new_field.input_type = self.input_type
        new_field.full_clean = self.full_clean
        return new_field

    @property
    def arguments(self):
        arguments = super().arguments
        if not self.input_type:
            return arguments

        return [
            *arguments,
            argument(
                self.argument_name,
                self.input_type,
            ),
        ]

    @arguments.setter
    def arguments(self, value: list[StrawberryArgument]):
        args_prop = super(DjangoMutationBase, self.__class__).arguments
        return args_prop.fset(self, value)  # type: ignore

    def refetch(self, resolved: _T, *, info: Info | None) -> _T:
        if not DjangoOptimizerExtension.enabled.get() or info is None:
            return resolved

        if isinstance(resolved, list) and resolved:
            model = type(resolved[0])
            if issubclass(model, models.Model):
                original_order = {r.pk: i for i, r in enumerate(resolved)}
                resolved_qs = optimize(
                    model._default_manager.filter(pk__in=[r.pk for r in resolved]),
                    info=info,
                )
                # sort the resolved objects in the order they were given
                resolved = sorted(  # type: ignore
                    resolved_qs,
                    key=lambda r: original_order[r.pk],
                )
        elif isinstance(resolved, models.Model):
            model = type(resolved)
            resolved = optimize(
                model._default_manager.filter(pk=resolved.pk),
                info=info,
            ).get()

        return resolved


class DjangoCreateMutation(DjangoMutationCUD, StrawberryDjangoFieldFilters):
    @django_resolver
    @transaction.atomic
    def resolver(
        self,
        source: Any,
        info: Info | None,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> Any:
        assert info is not None

        data: list[Any] | Any = kwargs.get(self.argument_name)

        # Do not optimize anything while retrieving the object to create
        with DjangoOptimizerExtension.disabled():
            if self.is_list:
                assert isinstance(data, list)
                resolved = [
                    self.create(
                        resolvers.parse_input(info, vars(d), key_attr=self.key_attr),
                        info=info,
                    )
                    for d in data
                ]
            else:
                assert not isinstance(data, list)
                resolved = self.create(
                    resolvers.parse_input(info, vars(data), key_attr=self.key_attr)
                    if data is not None
                    else {},
                    info=info,
                )

        return self.refetch(resolved, info=info)

    def create(self, data: dict[str, Any], *, info: Info):
        model = self.django_model
        assert model is not None

        return resolvers.create(
            info,
            model,
            data,
            key_attr=self.key_attr,
            full_clean=self.full_clean,
        )


def get_vdata(data: Any) -> dict[str, Any]:
    return vars(data).copy() if data is not None else {}


def get_pk(
    data: dict[str, Any],
    *,
    key_attr: str | None = None,
) -> strawberry.ID | relay.GlobalID | Literal[UNSET] | None:  # type: ignore
    if key_attr is None:
        settings = strawberry_django_settings()
        key_attr = settings["DEFAULT_PK_FIELD_NAME"]

    pk = data.pop(key_attr, UNSET)

    if pk is UNSET:
        pk = data.pop("id", UNSET)
    return pk


class DjangoUpdateMutation(DjangoMutationCUD, StrawberryDjangoFieldFilters):
    @django_resolver
    @transaction.atomic
    def resolver(
        self,
        source: Any,
        info: Info | None,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> Any:
        assert info is not None

        data: list[Any] | Any = kwargs.get(self.argument_name)

        # Do not optimize anything while retrieving the object to update
        with DjangoOptimizerExtension.disabled():
            if isinstance(data, list):
                resolved = [self.instance_level_update(info, kwargs, d) for d in data]
            else:
                resolved = self.instance_level_update(info, kwargs, data)

        return self.refetch(resolved, info=info)

    def instance_level_update(
        self,
        info: Info,
        kwargs: dict[str, Any],
        data: Any,
    ) -> Any:
        model = self.django_model
        assert model is not None

        vdata = get_vdata(data)
        pk = get_pk(vdata, key_attr=self.key_attr)

        if pk not in (None, UNSET):  # noqa: PLR6201
            instance = get_with_perms(
                pk,
                info,
                required=True,
                model=model,
                key_attr=self.key_attr,
            )
        else:
            instance = filter_with_perms(
                self.get_queryset(
                    queryset=model._default_manager.all(),
                    info=info,
                    **kwargs,
                ),
                info,
            )

        return self.update(
            info, instance, resolvers.parse_input(info, vdata, key_attr=self.key_attr)
        )

    def update(
        self,
        info: Info,
        instance: models.Model | Iterable[models.Model],
        data: dict[str, Any],
    ):
        return resolvers.update(
            info,
            instance,
            data,
            key_attr=self.key_attr,
            full_clean=self.full_clean,
        )


class DjangoDeleteMutation(
    DjangoMutationCUD,
    DjangoMutationBase,
    StrawberryDjangoFieldFilters,
):
    @django_resolver
    @transaction.atomic
    def resolver(
        self,
        source: Any,
        info: Info | None,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> Any:
        assert info is not None

        model = self.django_model
        assert model is not None

        data: Any = kwargs.get(self.argument_name)
        vdata = get_vdata(data)

        pk = get_pk(vdata, key_attr=self.key_attr)
        if pk not in (None, UNSET):  # noqa: PLR6201
            instance = get_with_perms(
                pk,
                info,
                required=True,
                model=model,
                key_attr=self.key_attr,
            )
        else:
            instance = filter_with_perms(
                self.get_queryset(
                    queryset=model._default_manager.all(),
                    info=info,
                    **kwargs,
                ),
                info,
            )

        return self.delete(
            info,
            instance,
            resolvers.parse_input(info, vdata, key_attr=self.key_attr),
        )

    def delete(
        self,
        info: Info,
        instance: models.Model | Iterable[models.Model],
        data: dict[str, Any] | None = None,
    ):
        return resolvers.delete(
            info,
            instance,
            data=data,
        )



import sys
import warnings
from typing import Generic, Optional, TypeVar, Union, cast

import strawberry
from django.db import DEFAULT_DB_ALIAS
from django.db.models import Count, QuerySet, Window
from django.db.models.functions import RowNumber
from django.db.models.query import MAX_GET_RESULTS  # type: ignore
from strawberry.types import Info
from strawberry.types.arguments import StrawberryArgument
from strawberry.types.unset import UNSET, UnsetType
from typing_extensions import Self

from strawberry_django.fields.base import StrawberryDjangoFieldBase
from strawberry_django.resolvers import django_resolver

from .arguments import argument
from .settings import strawberry_django_settings

NodeType = TypeVar("NodeType")
_QS = TypeVar("_QS", bound=QuerySet)

PAGINATION_ARG = "pagination"


@strawberry.type
class OffsetPaginationInfo:
    offset: int = 0
    limit: Optional[int] = UNSET


@strawberry.input
class OffsetPaginationInput(OffsetPaginationInfo): ...


@strawberry.type
class OffsetPaginated(Generic[NodeType]):
    queryset: strawberry.Private[Optional[QuerySet]]
    pagination: strawberry.Private[OffsetPaginationInput]

    @strawberry.field
    def page_info(self) -> OffsetPaginationInfo:
        return OffsetPaginationInfo(
            limit=self.pagination.limit,
            offset=self.pagination.offset,
        )

    @strawberry.field(description="Total count of existing results.")
    @django_resolver
    def total_count(self) -> int:
        return self.get_total_count()

    @strawberry.field(description="List of paginated results.")
    @django_resolver
    def results(self) -> list[NodeType]:
        paginated_queryset = self.get_paginated_queryset()

        return cast(
            "list[NodeType]",
            paginated_queryset if paginated_queryset is not None else [],
        )

    @classmethod
    def resolve_paginated(
        cls,
        queryset: QuerySet,
        *,
        info: Info,
        pagination: Optional[OffsetPaginationInput] = None,
        **kwargs,
    ) -> Self:
        """Resolve the paginated queryset.

        Args:
            queryset: The queryset to be paginated.
            info: The strawberry execution info resolve the type name from.
            pagination: The pagination input to be applied.
            kwargs: Additional arguments passed to the resolver.

        Returns:
            The resolved `OffsetPaginated`

        """
        return cls(
            queryset=queryset,
            pagination=pagination or OffsetPaginationInput(),
        )

    def get_total_count(self) -> int:
        """Retrieve tht total count of the queryset without pagination."""
        return get_total_count(self.queryset) if self.queryset is not None else 0

    def get_paginated_queryset(self) -> Optional[QuerySet]:
        """Retrieve the queryset with pagination applied.

        This will apply the paginated arguments to the queryset and return it.
        To use the original queryset, access `.queryset` directly.
        """
        from strawberry_django.optimizer import is_optimized_by_prefetching

        if self.queryset is None:
            return None

        return (
            self.queryset._result_cache  # type: ignore
            if is_optimized_by_prefetching(self.queryset)
            else apply(self.pagination, self.queryset)
        )


def apply(
    pagination: Optional[object],
    queryset: _QS,
    *,
    related_field_id: Optional[str] = None,
) -> _QS:
    """Apply pagination to a queryset.

    Args:
    ----
        pagination: The pagination input.
        queryset: The queryset to apply pagination to.
        related_field_id: The related field id to apply pagination to.
          When provided, the pagination will be applied using window functions
          instead of slicing the queryset.
          Useful for prefetches, as those cannot be sliced after being filtered

    """
    if pagination in (None, strawberry.UNSET):  # noqa: PLR6201
        return queryset

    if not isinstance(pagination, OffsetPaginationInput):
        raise TypeError(f"Don't know how to resolve pagination {pagination!r}")

    if related_field_id is not None:
        queryset = apply_window_pagination(
            queryset,
            related_field_id=related_field_id,
            offset=pagination.offset,
            limit=pagination.limit,
        )
    else:
        start = pagination.offset
        limit = pagination.limit
        if limit is UNSET:
            settings = strawberry_django_settings()
            limit = settings["PAGINATION_DEFAULT_LIMIT"]

        if limit is not None and limit >= 0:
            stop = start + limit
            queryset = queryset[start:stop]
        else:
            queryset = queryset[start:]

    return queryset


class _PaginationWindow(Window):
    """Window function to be used for pagination.

    This is the same as django's `Window` function, but we can easily identify
    it in case we need to remove it from the queryset, as there might be other
    window functions in the queryset and no other way to identify ours.
    """


def apply_window_pagination(
    queryset: _QS,
    *,
    related_field_id: str,
    offset: int = 0,
    limit: Optional[int] = UNSET,
) -> _QS:
    """Apply pagination using window functions.

    Useful for prefetches, as those cannot be sliced after being filtered.

    This is based on the same solution that Django uses, which was implemented
    in https://github.com/django/django/pull/15957

    Args:
    ----
        queryset: The queryset to apply pagination to.
        related_field_id: The related field id to apply pagination to.
        offset: The offset to start the pagination from.
        limit: The limit of items to return.

    """
    order_by = [
        expr
        for expr, _ in queryset.query.get_compiler(
            using=queryset._db or DEFAULT_DB_ALIAS  # type: ignore
        ).get_order_by()
    ]

    queryset = queryset.annotate(
        _strawberry_row_number=_PaginationWindow(
            RowNumber(),
            partition_by=related_field_id,
            order_by=order_by,
        ),
        _strawberry_total_count=_PaginationWindow(
            Count(1),
            partition_by=related_field_id,
        ),
    )

    if offset:
        queryset = queryset.filter(_strawberry_row_number__gt=offset)

    if limit is UNSET:
        settings = strawberry_django_settings()
        limit = settings["PAGINATION_DEFAULT_LIMIT"]

    # Limit == -1 means no limit. sys.maxsize is set by relay when paginating
    # from the end to as a way to mimic a "not limit" as well
    if limit is not None and limit >= 0 and limit != sys.maxsize:
        queryset = queryset.filter(_strawberry_row_number__lte=offset + limit)

    return queryset


def remove_window_pagination(queryset: _QS) -> _QS:
    """Remove pagination window functions from a queryset.

    Utility function to remove the pagination `WHERE` clause added by
    the `apply_window_pagination` function.

    Args:
    ----
        queryset: The queryset to apply pagination to.

    """
    queryset = queryset._chain()  # type: ignore
    queryset.query.where.children = [
        child
        for child in queryset.query.where.children
        if (not hasattr(child, "lhs") or not isinstance(child.lhs, _PaginationWindow))
    ]
    return queryset


def get_total_count(queryset: QuerySet) -> int:
    """Get the total count of a queryset.

    Try to get the total count from the queryset cache, if it's optimized by
    prefetching. Otherwise, fallback to the `QuerySet.count()` method.
    """
    from strawberry_django.optimizer import is_optimized_by_prefetching

    if is_optimized_by_prefetching(queryset):
        results = queryset._result_cache  # type: ignore

        if results:
            try:
                return results[0]._strawberry_total_count
            except AttributeError:
                warnings.warn(
                    (
                        "Pagination annotations not found, falling back to QuerySet resolution. "
                        "This might cause n+1 issues..."
                    ),
                    RuntimeWarning,
                    stacklevel=2,
                )

        # If we have no results, we can't get the total count from the cache.
        # In this case we will remove the pagination filter to be able to `.count()`
        # the whole queryset with its original filters.
        queryset = remove_window_pagination(queryset)

    return queryset.count()


class StrawberryDjangoPagination(StrawberryDjangoFieldBase):
    def __init__(self, pagination: Union[bool, UnsetType] = UNSET, **kwargs):
        self.pagination = pagination
        super().__init__(**kwargs)

    def __copy__(self) -> Self:
        new_field = super().__copy__()
        new_field.pagination = self.pagination
        return new_field

    def _has_pagination(self) -> bool:
        if isinstance(self.pagination, bool):
            return self.pagination

        if self.is_paginated:
            return True

        django_type = self.django_type
        if django_type is not None and not issubclass(
            django_type, strawberry.relay.Node
        ):
            return django_type.__strawberry_django_definition__.pagination

        return False

    @property
    def arguments(self) -> list[StrawberryArgument]:
        arguments = []
        if self.base_resolver is None and (self.is_list or self.is_paginated):
            pagination = self.get_pagination()
            if pagination is not None:
                arguments.append(
                    argument("pagination", OffsetPaginationInput, is_optional=True),
                )
        return super().arguments + arguments

    @arguments.setter
    def arguments(self, value: list[StrawberryArgument]):
        args_prop = super(StrawberryDjangoPagination, self.__class__).arguments
        return args_prop.fset(self, value)  # type: ignore

    def get_pagination(self) -> Optional[type]:
        return OffsetPaginationInput if self._has_pagination() else None

    def apply_pagination(
        self,
        queryset: _QS,
        pagination: Optional[object] = None,
        *,
        related_field_id: Optional[str] = None,
    ) -> _QS:
        return apply(pagination, queryset, related_field_id=related_field_id)

    def get_queryset(
        self,
        queryset: _QS,
        info: Info,
        *,
        pagination: Optional[OffsetPaginationInput] = None,
        _strawberry_related_field_id: Optional[str] = None,
        **kwargs,
    ) -> _QS:
        queryset = super().get_queryset(queryset, info, **kwargs)

        # This is counter intuitive, but in case we are returning a `Paginated`
        # result, we want to set the original queryset _as is_ as it will apply
        # the pagination later on when resolving its `.results` field.
        # Check `get_wrapped_result` below for more details.
        if self.is_paginated:
            return queryset

        # Add implicit pagination if this field is not a list
        # that way when first() / get() is called on the QuerySet it does not cause extra queries
        # and we don't prefetch more than necessary
        if not pagination and not (
            self.is_list or self.is_paginated or self.is_connection
        ):
            if self.is_optional:
                # first() applies order by pk if not ordered
                if not queryset.ordered:
                    queryset = queryset.order_by("pk")
                pagination = OffsetPaginationInput(offset=0, limit=1)
            else:
                pagination = OffsetPaginationInput(offset=0, limit=MAX_GET_RESULTS)

        return self.apply_pagination(
            queryset,
            pagination,
            related_field_id=_strawberry_related_field_id,
        )
